/* 
   Exercise on the use of FFTW package
   written by Pierluigi Monaco, Trieste University, March 2017
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fftw3.h>

#define PI      3.14159265358979323846

int main(int argc, char **argv)
{

  /* arguments from the command line */
  if (argc<3)
    {
      printf("Usage: fft.x grid derivative\n");
      return 0;
    }

  printf("Starting...\n");

  /* grid size */
  int M = atoi(argv[1]);

  /* derivative */
  int my_derivative = atoi(argv[2]);
  if (my_derivative < 0 || my_derivative > 2)
    {
      printf("Error: invalid value for derivative\n");
      return 0;
    }

  if (my_derivative==0)
    printf("I will return the starting field\n");
  else if (my_derivative==1)
    printf("I will return the x-derivative of the field\n");
  else if (my_derivative==2)
    printf("I will return the y-derivative of the field\n");

  
  fftw_plan forward_plan, reverse_plan;
  fftw_complex *real_space, *fourier_space;
  int ndata, i,j,ii,n;

  /* size of vectors to allocate */
  ndata=M*M;

  /* allocation of vectors for ffts */
  real_space = fftw_alloc_complex(ndata);
  fourier_space = fftw_alloc_complex(ndata);

  if (real_space == 0x0 || fourier_space == 0x0)
    {
      printf("ERROR: I cannot allocate memory for fft vectors\n");
      fflush(stdout);
      return 1;
    }

  /* construction of fftw plans */
  forward_plan = fftw_plan_dft_2d(M, M,                       // grid size
				  real_space, fourier_space,  // vectors to transform
				  FFTW_FORWARD,               // sign of transform => -1
				  FFTW_MEASURE);              // fftw parameters
  reverse_plan = fftw_plan_dft_2d(M, M,                       // grid size
				  fourier_space, real_space,  // vectors to transform
				  FFTW_BACKWARD,              // sign of transform => +1
				  FFTW_MEASURE); 	      // fftw parameters

  /* reading in the real space field to transform */
  double *data = (double*)malloc(ndata * sizeof(double));

  FILE *fd = fopen("field.dat","r");
  if (fd == 0x0)
    {
      printf("ERROR: file field.dat not found\n");
      return 1;
    }
  fread(data, sizeof(double), ndata, fd);
  fclose(fd);

  for (i=0; i<ndata; i++)
    {
      real_space[i][0]=data[i];
      real_space[i][1]=0.0;
    }

  /* FFT forward transform */
  fftw_execute(forward_plan);

  printf("Forward transform done\n");

  /* Writes the transform on the file */
  fd = fopen("transform.dat","w");
  fwrite(fourier_space, sizeof(double), 2*ndata, fd);
  fclose(fd);

  printf("Written fourier-space field in file transform.dat\n");

  if (my_derivative>0)
    {

      double kx, ky,tmp;
      int jj, location;
      
      for (i=0; i<M; i++)
	{
	  ii = i;
	  if (ii > M/2)
	    ii -= M;
	  kx = 2.0 * PI * ii;

	  for (j=0; j<M; j++)
	    {
	      jj = j;
	      if (j > M/2)
		jj -= M;
	      ky = 2.0 * PI * jj;

	      location = i*M + j;
	      
	      tmp = fourier_space[location][0];
	      if (my_derivative==1)
		{
		  fourier_space[location][0] = - kx * fourier_space[location][1];
		  fourier_space[location][1] =   kx * tmp;
		}
	      else
		{
		  fourier_space[location][0] = - ky * fourier_space[location][1];
		  fourier_space[location][1] =   ky * tmp;
		}
	    }
	}

    }

  
  /* Test of backward transform */
  fftw_execute(reverse_plan);

  double norm = 1.0/(double)(M*M);
  for (i=0; i<ndata; i++)
    {
      real_space[i][0] *= norm;
      real_space[i][1] *= norm;
    }

  /* Check on the imaginary part */
  double sum=0.0;
  for (i=0; i<ndata; i++)
    sum+=fabs(real_space[i][1]);
  sum/=(double)(ndata);
  printf("Average of the modulus of imaginary part: %g\n",sum);

  /* write out the result */
  char filename[50];
  if (my_derivative==0)
    strcpy(filename,"back.dat");
  else if (my_derivative==1)
    strcpy(filename,"x_derivative.dat");
  else if (my_derivative==2)
    strcpy(filename,"y_derivative.dat");
  
  /* Stores the real part in the data vector */
  for (i=0; i<ndata; i++)
    data[i]=real_space[i][0];

  fd = fopen(filename,"w");
  fwrite(data, sizeof(double), ndata, fd);
  fclose(fd);

  printf("Written file %s\n",filename);

  free(data);

  fftw_destroy_plan(reverse_plan);
  fftw_destroy_plan(forward_plan);

  return 0;

}
